////////////////////////////////////////////////////////////////
// j_ace.h
////////////////////////////////////////////////////////////////
// J-ACE (Jaka Arcade Controls Encoder) for HW v.1.0
// Copyright 2020 Jaakko Kairus
// jaakko@kair.us
//
// You may freely use and modify this software for personal use.
// Commercial use of this software or parts of it is prohibited
// without written permission from author.
// DISCLAIMER: This software comes with ABSOLUTELY NO WARRANTY
//
// For more information about this project see project web page at
// http://kair.us/projects/j_ace/

#define FW_VER   _fw08
#define HW_VER   _hw10
#define CAT(a,b,c,d) a ## b ## c ## d
#define EXPNAME(a,b,c,d) CAT(a,b,c,d)

#include <18F24K50.h>
#device ADC=8
#build(stack=256)
#fuses PLL3X,INTRC,NOWDT,NOFCMEN,NOIESO,BROWNOUT,BORV28,NOPROTECT,NOLVP,NODEBUG,PLLEN,NOCPUDIV,PUT,NOMCLR
#use delay(int=16MHz, clock=48MHz, USB_FULL, act=USB)
#use fast_io(c)

#include <pic18_usb.h>					// Microchip PIC18Fxx5x hardware layer for usb.c
//#include <usb_desc_j_ace_singlekbd.h>	// USB Configuration and Device descriptors for a single HID keyboard
//#include <usb_desc_j_ace_nkro.h>    	// USB Configuration and Device descriptors for a bitmap based NKRO keyboard which also includes boot protocol keyboard for compatibility
#ifdef J_ACE_JOYSTICK
	#include <usb_desc_j_ace_joystick.h>	// USB descriptors for joystick
	#export (HEX, FILE = EXPNAME(j-ace_joystick,HW_VER,FW_VER,.hex)) 
#else
	#include <usb_desc_j_ace_trplkbd.h>   	// USB descriptors for three standard HID keyboards which together allow NKRO with J-ACE
	#export (HEX, FILE = EXPNAME(j-ace_keyboard,HW_VER,FW_VER,.hex)) 
#endif
#include <usb.c>						// handles usb setup tokens and get descriptor reports
#include <stdlib.h>
#include "usb_bootloader.h"				// sets program location and reset vector for use with bootloader
#include "usb_hid_keys.h"				// keyboard scancodes
#import(FILE=bootloader-PIC18F24K50-dav-232.hex,HEX,RANGE=0:0x7FF)		// Embed the bootloader binary

#define PRESS_DEBOUNCE_MS 20			// Input debounce time when switch is closed (button pressed down)
#define RELEASE_DEBOUNCE_MS 10			// Debounce time when switch is opened (button released)

#define AUTOFIRE_DELAY_LOW_LIMIT 8		// Autofire speed high limit. 8 ms x 2 equals 62.5 Hz (62 button clicks per second!)
#define AUTOFIRE_DELAY_HIGH_LIMIT 128	// Autofire speed low limit. 128 ms x 2 equals ca. 4 Hz (4 button clicks per second)
#define AUTOFIRE_DEFAULT_DELAY 32		// Autofire default speed is 15 cliks per second.

#define TOTAL_DEBOUNCE_MS PRESS_DEBOUNCE_MS+RELEASE_DEBOUNCE_MS

#if TOTAL_DEBOUNCE_MS > 255
#error Too long debounce time, max total length is 255 ms
#endif

// Input pin defines. Most of these are currently unused since inputs are read as whole port
#define PIN_BUT_1     PIN_A0
#define PIN_BUT_2     PIN_A1
#define PIN_BUT_3     PIN_A2
#define PIN_BUT_4     PIN_A3
#define PIN_BUT_5     PIN_A4
#define PIN_BUT_6     PIN_A5
#define PIN_BUT_7     PIN_A7
#define PIN_BUT_8     PIN_A6
#define PIN_JOY_UP    PIN_B0
#define PIN_JOY_DOWN  PIN_B1
#define PIN_JOY_LEFT  PIN_B2
#define PIN_JOY_RIGHT PIN_B3
#define PIN_BUT_START PIN_B4
#define PIN_BUT_COIN  PIN_B5
#define PIN_BUT_A     PIN_B6
#define PIN_BUT_B     PIN_B7
#define PIN_AUTOFIRE  PIN_C2
#define PIN_LED_1     PIN_C0
#define PIN_LED_2     PIN_C1
#define PIN_LED_3     PIN_C6
#define PIN_LED_4     PIN_C7

// Keyboard scan code map for Player 1, 2, 3 and 4 button definitions
// The Shift array is used for Player 1 when start button is held down
// The Mod parameter must be set to 1 if the assigned scan code is a modifier
// key (Alt, Control, Shift or Meta key)
unsigned int8 const kbd_scancodes[16][10] = {
//	Player 1		Mod	Player 2	Mod Player 3		Mod	Player 4	Mod	Shift			Mod    Input
	KEY_MOD_LCTRL,	1,	KEY_A,		0,	KEY_MOD_RCTRL,	1,	KEY_B,		0,	KEY_5,			0,	// Button 1
	KEY_MOD_LALT,	1,	KEY_S,		0,	KEY_MOD_RSHIFT,	1,	KEY_E,		0,	KEY_ESC,		0,	// Button 2
	KEY_SPACE,		0,	KEY_Q,		0,	KEY_ENTER,		0,	KEY_H,		0,	KEY_SPACE,		0,	// Button 3
	KEY_MOD_LSHIFT,	1,	KEY_W,		0,	KEY_O,			0,	KEY_M,		0,	KEY_MOD_LSHIFT,	1,	// Button 4
	KEY_Z,			0,	KEY_I,		0,	KEY_F1,			0,	KEY_F7,		0,	KEY_Z,			0,	// Button 5
	KEY_X,			0,	KEY_K,		0,	KEY_F2,			0,	KEY_F8,		0,	KEY_X,			0,	// Button 6
	KEY_V,			0,	KEY_L,		0,	KEY_F4,			0,	KEY_F10,	0,	KEY_V,			0,	// Button 8		Note the order of
	KEY_C,			0,	KEY_J,		0,	KEY_F3,			0,	KEY_F9,		0,	KEY_C,			0,	// Button 7		buttons 7 and 8!
	KEY_UP,			0,	KEY_R,		0,	KEY_I,			0,	KEY_Y,		0,	KEY_GRAVE,		0,	// Joystick up
	KEY_DOWN,		0,	KEY_F,		0,	KEY_K,			0,	KEY_N,		0,	KEY_P,			0,	// Joystick down
	KEY_LEFT,		0,	KEY_D,		0,	KEY_J,			0,	KEY_V,		0,	KEY_ENTER,		0,	// Joystick left
	KEY_RIGHT,		0,	KEY_G,		0,	KEY_L,			0,	KEY_U,		0,	KEY_TAB,		0,	// Joystick right
	KEY_1,			0,	KEY_2,		0,	KEY_3,			0,	KEY_4,		0,	KEY_1,			0,	// Start button
	KEY_5,			0,	KEY_6,		0,	KEY_7,			0,	KEY_8,		0,	KEY_5,			0,	// Coin acceptor / button
	KEY_P,			0,	KEY_TAB,	0,	KEY_F5,			0,	KEY_F11,	0,	KEY_P,			0,	// Button A
	KEY_ENTER,		0,	KEY_ESC,	0,	KEY_F6,			0,	KEY_F12,	0,	KEY_ENTER,		0	// Button B
};

// Index defines for button arrays
#define IDX_BUT_1     0
#define IDX_BUT_2     1
#define IDX_BUT_3     2
#define IDX_BUT_4     3
#define IDX_BUT_5     4
#define IDX_BUT_6     5
#define IDX_BUT_7     7
#define IDX_BUT_8     6
#define IDX_JOY_UP    8
#define IDX_JOY_DOWN  9
#define IDX_JOY_LEFT  10
#define IDX_JOY_RIGHT 11
#define IDX_BUT_START 12
#define IDX_BUT_COIN  13
#define IDX_BUT_A     14
#define IDX_BUT_B     15
#define IDX_AUTOFIRE  16

// Global variables
unsigned int8 debounce_timer[17]={0};
unsigned int8 autofire_speed_setting = 32;		// autofire delay is ms. Total cycle is 2 x setting. Thus this default is 15.625 Hz
unsigned int8 player = 0;						// 0 for player 1, 2 for player 2, 4 for player 3, 6 for player 4
unsigned int8 shift_active = 0;
int1 autofire_buttons[16]={0};
int1 previous_buttons[17]={0};
int1 current_startbutton = 0;
int1 millisec_passed = 0;

struct parts {
   unsigned int8 a;
   unsigned int8 b;
   unsigned int1 c;
};

union {
	int1 bits[17];
	struct parts bytes;
} active_buttons;							// Debounced button data
	
union {
	int1 bits[17];
	struct parts bytes;
} output_buttons;							// The data to be sent via USB


// Function prototypes
void add_autofire();
void send_keycodes_trplkbd();
void handle_shift();
void read_inputs();
void read_debounced_inputs();
void load_jace_settings();
void setup_keymap();
void usb_joystick_task(void);
void measure_millisec();